#define SPR_ATTR_ENABLE		0x01
#define SPR_ATTR_FLIP_H		0x08
#define SPR_ATTR_FLIP_V		0x02

#define TIA91_CPU_FREQ      (15750000 / 9)
#define TIA91_TICKS_VSYNC   57
#define TIA91_TICKS_FRAME   (TIA91_CPU_FREQ / 50)

namespace EmuTIA91
{
const double g_v[] = { 1.2071, 0.9971, 0.9259, 0.7159, 0.4912, 0.2812, 0.2100, 0.0000 };
const double r_v[] = { 1.5937, 1.3125, 1.1562, 0.8750, 0.7187, 0.4375, 0.2812, 0.0000 };
const double b_v[] = { 1.3523, 0.8750, 0.4773, 0.0000 };

#ifdef PALETTE_R8G8B8
unsigned int palette888[256];
#endif
#ifdef PALETTE_R3G3B2
unsigned char palette332[256];
#endif

unsigned char* pmem = 0;		//0000..dfff - ПЗУ графики фона и кода, c000..cfff - основное ОЗУ, f000..f7ff карта тайлов
unsigned char* srom = 0;		//0000..7fff - ПЗУ графики спрайтов
unsigned char* crom = 0;		//0000..7fff - ПЗУ графики спрайтов

unsigned char vpal[16];			//палитра

unsigned char spr_x[16];
unsigned char spr_y[16];
unsigned char spr_n[16];
unsigned char spr_a[16];

unsigned char pio_d0;
unsigned char pio_d1;
unsigned char pio_d2;

unsigned char vsync;

unsigned char vram_sel;

unsigned char scr_xoff;
unsigned char scr_yoff;

unsigned char port_rd(unsigned char adr)
{
  switch (adr)
  {
    case 0xD0: return pio_d0;
    case 0xD1: return pio_d1;
    case 0xD2: return pio_d2 | vsync;
  }

  return 0;
}

void port_wr(unsigned char adr, unsigned char val)
{
  switch (adr & 0xf0)
  {
    case 0x00:
      spr_y[adr - 0x00] = val ^ 0xff;	//sprite Y
      return;

    case 0x10:
      spr_x[adr - 0x10] = val ^ 0xff;	//sprite X
      return;

    case 0x20:
      spr_n[adr - 0x20] = val ^ 0xff;	//sprite num
      return;

    case 0x30:
      spr_a[adr - 0x30] = val ^ 0xff;	//sprite attr
      return;

    case 0xe0:	//palette
      adr -= 0xe0;
      vpal[adr] = palette332[val];
      return;
  }

  switch (adr)	//everything else
  {
    case 0xf0://vertical scroll
      //scr_yoff = ((val - 1) ^ 0xff) & 0xff;
      return;

    case 0xf4://horizontal scroll
      //scr_xoff = ((val - 6) ^ 0xff) & 0xff;
      return;

    case 0xf8://character bitplanes, video page select
      vram_sel = val;
      return;

    case 0xc0://PIT ch 0
    case 0xc1://PIT ch 1
    case 0xc2://PIT ch 2
    case 0xc3://PIT control word

      SoundTIA::write(CPU8080::time, adr, val);

    case 0xd2://coin register control
    case 0xd3://PPI control word
      return;
  }
}

void mem_wr(unsigned short int adr, unsigned char val)
{
  if (adr >= 0xc000)
  {
    pmem[adr] = val;

    return;
  }
}

unsigned char mem_rd(unsigned short int adr)
{
  return pmem[adr];
}

int update(int pad_state, bool audio_underrun)
{
  pio_d0 = 0;
  pio_d1 = 0;
  pio_d2 = 0;

  if (pad_state & PAD_LEFT)  pio_d0 |= 0x20;
  if (pad_state & PAD_RIGHT) pio_d0 |= 0x02;
  if (pad_state & PAD_UP)    pio_d1 |= 0x02;
  if (pad_state & PAD_DOWN)  pio_d1 |= 0x20;

  if (pad_state & PAD_TEST)  pio_d1 |= 0x80;
  if (pad_state & PAD_COIN)  pio_d2 |= 0x10;
  if (pad_state & PAD_A)     pio_d2 |= 0x40;
  if (pad_state & PAD_B)     pio_d2 |= 0x20;

  pio_d0 ^= 0x22;
  pio_d1 ^= 0x22 | 0x80;
  pio_d2 ^= 0x20 | 0x40;

  vsync |= 0x80;

  CPU8080::time = 0;

  CPU8080::run(TIA91_TICKS_VSYNC);

  vsync &= ~0x80;

  int t = TIA91_TICKS_FRAME - TIA91_TICKS_VSYNC;

  if (audio_underrun) t += TIA91_TICKS_FRAME / 10;

  CPU8080::run(t);

  SoundTIA::write(CPU8080::time, 0, 0);	//mark end of frame

  return t;
}

void render_audio(int t)
{
  SoundTIA::render(t);
}

void render_tile(unsigned char** raster, int tile, int xs, int ys)
{
  int ptr = (((vram_sel & 0xfe) << 7) + (tile << 3)) & 0x1ff8;

  for (int y = 0; y < 8; ++y)
  {
    int sy = ((ys + y) & 255);

    if (sy >= 8 && sy < 248)
    {
      sy -= 8;
      int mask = 0x80;

      for (int x = 0; x < 8; ++x)
      {
        unsigned char col = (crom[ptr + 0x0000] & mask ? 1 : 0) | (crom[ptr + 0x2000] & mask ? 2 : 0) | (crom[ptr + 0x4000] & mask ? 4 : 0) | (crom[ptr + 0x6000] & mask ? 8 : 0);

        mask >>= 1;

        raster[sy][(xs + x) & 255] = vpal[col];
      }
    }

    ++ptr;
  }
}

void render_sprite(unsigned char** raster, int spr, int xs, int ys, bool flip_h, bool flip_v)
{
  int x_flip = 0;
  int y_flip = 0;

  if (flip_h) x_flip = 15;
  if (flip_v) y_flip = 15;

  int ptr = spr << 4;

  for (int y = 0; y < 16; ++y)
  {
    int sy = ((ys + (y ^ y_flip)) & 255);

    if (sy >= 8 && sy < 248)
    {
      sy -= 8;

      for (int x = 0; x < 16; ++x)
      {
        int ptrx = ptr + ((x & 8) << 9);
        int mask = 128 >> (x & 7);

        unsigned char col = (srom[ptrx + 0x0000] & mask ? 1 : 0) | (srom[ptrx + 0x2000] & mask ? 2 : 0) | (srom[ptrx + 0x4000] & mask ? 4 : 0) | (srom[ptrx + 0x6000] & mask ? 8 : 0);

        if (col < 0x0f)
        {
          raster[sy][(xs + (x ^ x_flip)) & 255] = vpal[col];
        }
      }
    }

    ++ptr;
  }
}

void render_video(unsigned char** raster)
{
  //render BG

  //if (!(vram_sel & 0x80)) vram_ptr = 0xf000; else vram_ptr = 0xf400;
  int vram_ptr = 0xf000;

  int y = scr_yoff;

  for (int i = 0; i < 32; ++i)
  {
    int x = scr_xoff;

    for (int j = 0; j < 32; ++j)
    {
      render_tile(raster, pmem[vram_ptr++], x, y);

      x += 8;
    }

    y += 8;
  }

  //render sprites

  for (int id = 0; id < 16; ++id)
  {
    if (!(spr_a[id] & SPR_ATTR_ENABLE)) continue;

    render_sprite(raster, spr_n[id], spr_x[id], spr_y[id], spr_a[id] & SPR_ATTR_FLIP_H ? true : false, spr_a[id] & SPR_ATTR_FLIP_V ? true : false);
  }
}

void srom_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &srom[adr]);
}

void pmem_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &pmem[adr]);
}

void crom_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &crom[adr]);
}

void init(void)
{
  pmem = global_ram1; //65536
  srom = global_ram2; //32768
  crom = global_ram2 + 32768; //32768

  SoundTIA::init(2);

  vsync = 0;
  vram_sel = 0;

  scr_xoff = 0;
  scr_yoff = 0;

  pio_d0 = 0;
  pio_d1 = 0;
  pio_d2 = 0;

  for (int i = 0; i < 16; ++i) spr_a[i] = 0;

  //make palette

  for (int i = 0; i < 256; ++i)
  {
    int ir = (i >> 3) & 7;
    int ig = i & 7;
    int ib = (i >> 6) & 3;

    double dr = 255.0 * r_v[ir] / r_v[0];
    double dg = 255.0 * g_v[ig] / g_v[0];
    double db = 255.0 * b_v[ib] / b_v[0];

    int r = (255 - (int)dr) & 255;
    int g = (255 - (int)dg) & 255;
    int b = (255 - (int)db) & 255;

#ifdef PALETTE_R8G8B8
    palette888[i] = (r << 16) | (g << 8) | b;
#endif
#ifdef PALETTE_R3G3B2
    palette332[i] = ((r >> 5) << 5) | ((g >> 5) << 2) | (b >> 6);
#endif
  }

  CPU8080::init(mem_rd, mem_wr, port_rd, port_wr);
}

void reset(void)
{
  CPU8080::reset();
  SoundTIA::reset();
}
}
